import os
import pwd
import subprocess
import re
import glob
from jinja2 import Template
import shutil
import time
from hashlib import sha1
from urllib.parse import urlparse
from pathlib import Path

from charmhelpers import fetch
from charmhelpers.core import host, hookenv, unitdata
from charmhelpers.core.templating import render
from charmhelpers.contrib.charmsupport import nrpe
from charms.reactive import (
    when, when_not, set_state, remove_state, is_state, hook
)
from charms.reactive.helpers import any_file_changed, data_changed
from charms.layer import snap
import yaml


PATHS = {
    'prometheus_yml_tmpl': 'prometheus.yml.j2',
    'prometheus_def_tmpl': 'etc_default_prometheus.j2',
    'prometheus_rules_cron': '/etc/cron.d/juju-prometheus-rules-refresh',
    'prometheus_ssh_keyfile': '/root/.ssh/prometheus_git_key',
    'refresh_rules_script': '/usr/local/bin/refresh_rules.sh',
    'refresh_rules_script_tmpl': 'refresh_rules.sh.j2',
    'prometheus_yml': '/var/snap/prometheus/current/prometheus.yml',
    'prometheus_def': '/var/snap/prometheus/current/daemon_arguments',
    'prometheus_promreg_tgts_yml': '/var/snap/prometheus/current/promreg/targets.yaml',
    'custom_rules_path': '/var/snap/prometheus/current/custom.rules',
    'juju_rules_path': '/var/snap/prometheus/current/juju.rules',
    'external_rules_path': '/var/snap/prometheus/current/custom.rules.d',
    'promtool': '/snap/bin/prometheus.promtool',
    'service_name': 'snap.prometheus.prometheus',
    'promreg_yml': '/var/snap/promreg/common/promreg.yaml',
    'promreg_tgts_yml': '/var/snap/promreg/common/promreg/targets.yaml',
    'promreg_service_name': 'snap.promreg.promreg',
    'certs_dir': '/var/snap/prometheus/common/certs',
}


def templates_changed(tmpl_list):
    return any_file_changed(['templates/{}'.format(x) for x in tmpl_list])


def render_template(template, context):
    tmpl = Template(template, lstrip_blocks=True, trim_blocks=True)
    return tmpl.render(**context)


def filter_conflicting_daemon_arg(arg):
    return arg  # Just a stub function for now.


def parse_daemon_args_string(daemon_args):
    return list(filter(filter_conflicting_daemon_arg, daemon_args.split(";")))


def get_daemon_args_list_from_config():
    config = hookenv.config()
    daemon_args = config.get('daemon-args')
    assert isinstance(daemon_args, str)

    if daemon_args:
        hookenv.log("Retrieving {} from config".format(daemon_args),
                    level="DEBUG")
        return parse_daemon_args_string(daemon_args)
    else:
        hookenv.log("No daemon args in config", level="DEBUG")
        return []


# This could probably be a bit more intelligent.
# Curently we just return the processed list of
# string of args.
def daemon_args():
    args_list = get_daemon_args_list_from_config()
    return sorted(args_list)


def configure_rules():
    config = hookenv.config()
    # Static rules provided as config options
    if config.get('custom-rules', False):
        with open(PATHS['custom_rules_path'], 'w') as fh:
            fh.write(config['custom-rules'])

    kv = unitdata.kv()
    if config.get('rules-repo', False) or kv.get('prometheus.rules-repo') != config.get('rules-repo'):
        kv.set('prometheus.rules-repo', config.get('rules-repo', False))
        # Remove files on disk if rules-repo is empty or has changed
        remove_external_rules()
        set_state('prometheus.do-reconfig-yml')

    if config.get('rules-repo', False):
        ssh_command = ''
        if config.get('rules-update-ssh-key', False):
            host.mkdir(os.path.dirname(PATHS['prometheus_ssh_keyfile']), perms=0o700)
            with open(PATHS['prometheus_ssh_keyfile'], 'w') as f:
                f.write(config['rules-update-ssh-key'])
            os.chmod(PATHS['prometheus_ssh_keyfile'], 0o600)
            ssh_command = 'ssh -i {}'.format(PATHS['prometheus_ssh_keyfile'])

        rules_path = PATHS['external_rules_path']
        if os.path.exists(rules_path):
            subprocess.check_call(['git', '-C', rules_path, 'pull'],
                                  env={'GIT_SSH_COMMAND': ssh_command})
        else:
            # New repo, store key during initial clone
            subprocess.check_call(['git', 'clone', config['rules-repo'], rules_path],
                                  env={'GIT_SSH_COMMAND': ssh_command + ' -o StrictHostKeyChecking=no'})

        options = {
            'prometheus_yml': PATHS['prometheus_yml'],
            'external_rules_path': PATHS['external_rules_path'],
            'promtool': PATHS['promtool'],
            'ssh_command': ssh_command,
            'rules_update_notify_email': config.get('rules-update-notify-email', False),
        }
        render(source=PATHS['refresh_rules_script_tmpl'],
               target=PATHS['refresh_rules_script'],
               context=options,
               perms=0o755,
               )

        schedule = config.get('rules-update-schedule', '*/10 * * * *')
        with open(PATHS['prometheus_rules_cron'], 'w') as f:
            f.write('{} root {} &>/dev/null\n'.format(schedule, PATHS['refresh_rules_script']))
        set_state('prometheus.do-reconfig-yml')


def remove_external_rules():
    if os.path.exists(PATHS['prometheus_rules_cron']):
        os.unlink(PATHS['prometheus_rules_cron'])
    if os.path.exists(PATHS['prometheus_ssh_keyfile']):
        os.unlink(PATHS['prometheus_ssh_keyfile'])
    if os.path.exists(PATHS['external_rules_path']):
        shutil.rmtree(PATHS['external_rules_path'])


def get_rule_files():
    """Return list of rule files or globs

    We need to provide globs to handle situation where files disappear.
    At the same time filepath_glob parameter does not fully support globs,
    only filename part can be wildcarded.
    """
    config = hookenv.config()
    rule_files = []
    if config.get('custom-rules', False):
        rule_files.append(PATHS['custom_rules_path'])
    if os.path.isfile(PATHS['juju_rules_path']):
        rule_files.append(PATHS['juju_rules_path'])

    for root, dirs, filenames in os.walk(PATHS['external_rules_path']):
        for filename in filenames:
            match = re.match('^.*\.(yaml|yml|rules)$', filename)
            if match:
                # Wildcard filename part to allow rule files to disappear
                rule_files.append('{}/*.{}'.format(root, match.group(1)))
    return sorted(list(set(rule_files)))  # dedup entries


# TODO: once there's reactive support for storage hooks, convert off @hook()
@hook('metrics-filesystem-storage-attached')
def configure_storage():
    storage_path = subprocess.check_output(
        ['storage-get', 'location']).decode().strip()
    kv = unitdata.kv()
    kv.set('storage-path', storage_path)
    set_state('storage.configured')
    set_state('prometheus.do-check-reconfig')


@hook('install')
def install():
    set_state('prometheus.do-install')


def set_datadir_perms():
    datadir = unitdata.kv().get('storage-path', False)
    if not datadir:
        # No juju storage attached, use defaults from package
        return
    users = [i for i in pwd.getpwall() if i.pw_name == 'nobody']
    if len(users) == 1:
        os.lchown(datadir, users[0].pw_uid, users[0].pw_gid)


@when('prometheus.do-install')
def install_packages():
    hookenv.status_set('maintenance', 'Installing software')
    pkgname = 'prometheus'
    promreg_pkgname = 'promreg'
    config = hookenv.config()
    channel = config.get('snap_channel', 'stable')

    if not os.path.exists('/usr/sbin/sendmail'):
        fetch.apt_install('sendmail-bin')  # Needed for cronjob email notifications
    snap.install('core')  # Needed for offline deploy; falls back to store as needed
    snap.install(pkgname, channel=channel, force_dangerous=False)
    if config.get('prometheus_registration_listen', False):
        channel = config.get('promreg_snap_channel', 'stable')
        snap.install(promreg_pkgname, channel=channel, force_dangerous=False)
        subprocess.check_call(['snap', 'connect', 'prometheus:content',
                               'promreg:content'])
    remove_state('prometheus.do-install')
    set_state('prometheus.do-restart')
    hookenv.status_set('maintenance', 'Software installed')


def runtime_args(key=None, value=None):
    kv = unitdata.kv()
    args = kv.get('runtime_args', {})
    if key:
        args.update({key: value})
        kv.set('runtime_args', args)
    args_list = ['{} {}'.format(k, v) for k, v in args.items() if v] + \
        daemon_args()
    # sorted list is needed to avoid data_changed() false-positives
    return sorted(args_list)


def validate_config():
    subprocess.check_call([PATHS['promtool'], 'check', 'config', PATHS['prometheus_yml']])


def get_version():
    """Extract version info
    Defaults to 2.0.0
    """
    pattern = b'.*version (\d+)\.(\d+)\.(\d+) .*$'
    output = subprocess.check_output([PATHS['promtool'], '--version']).split(b'\n')[0]
    if re.match(pattern, output):
        major, minor, patch = re.match(pattern, output).groups()
        return (int(major), int(minor), int(patch))
    else:
        return (2, 0, 0)


def get_manual_jobs():
    config = hookenv.config()
    if not config.get('scrape-jobs', False):
        return None
    try:
        yaml.safe_load(config.get('scrape-jobs'))
    except Exception:
        return None
    snmp_exporter = unitdata.kv().get('snmp_exporter_address', '127.0.0.1:9116')
    blackbox_exporter = unitdata.kv().get('blackbox_exporter_address', '127.0.0.1:9115')
    t = Template(config.get('scrape-jobs'))
    return t.render(snmp_exporter=snmp_exporter, blackbox_exporter=blackbox_exporter)


@when('prometheus.do-reconfig-yml')
def write_prometheus_config_yml():
    hookenv.status_set('maintenance', 'Updating configuration')
    config = hookenv.config()

    ver_major, ver_minor, ver_patch = get_version()

    # transform eg. 'h1:p1 ,  h2:p2' (string), to ['h1:p1', 'h2:p2'] (list)
    static_targets = None
    if config.get('static-targets'):
        static_targets = [x.strip()
                          for x in config.get('static-targets', '').split(',')]
    federate_config = None
    if config.get('federate-config'):
        federate_config = yaml.safe_load(config.get('federate-config')) or None

    monitor_name = '{}-monitor'.format(hookenv.service_name())
    if config.get('monitor-name', False):
        monitor_name = config.get('monitor-name')

    # convert e.g. 'l1=v1,l2=v2' in {'l1': 'v1, 'l2'; 'v2}
    external_labels_config = config.get('external-labels', '').replace(' ', '')
    if external_labels_config:
        external_labels = dict(
            entry.split('=') for entry in external_labels_config.split(','))
    else:
        external_labels = {}

    prom_metrics_path = "/metrics"
    external_url = config.get("external_url", "")
    if external_url:
        ext_url = urlparse(external_url)
        prom_metrics_path = '{}/metrics'.format(ext_url.path.strip('/'))

    options = {
        'scrape_interval': config['scrape-interval'],
        'evaluation_interval': config['evaluation-interval'],
        'static_targets': static_targets,
        'private_address': hookenv.unit_get('private-address'),
        'listen_port': config['web-listen-port'],
        'promreg': config.get('prometheus_registration_listen', False),
        'promreg_tgts_yml': PATHS['promreg_tgts_yml'],
        'monitor_name': monitor_name,
        'prom_metrics_path': prom_metrics_path,
        'external_labels': external_labels,
        'rule_files': get_rule_files(),
        'prometheus_promreg_tgts_yml': PATHS['prometheus_promreg_tgts_yml'],
        'prometheus_registration_listen': config.get('prometheus_registration_listen', ''),
        'prometheus_registration_port': config.get('prometheus_registration_port', ''),
        'prometheus_registration_authtoken': config.get('prometheus_registration_authtoken', ''),
        'jobs': unitdata.kv().get('target_jobs', []),
        'federate': federate_config,
        'remote_read': config.get('remote-read', False),
        'remote_write': config.get('remote-write', False),
        'scrape_jobs': unitdata.kv().get('scrape_jobs', []),
        'rel_manual_jobs': unitdata.kv().get('manual_jobs', []),
        'cfg_manual_jobs': get_manual_jobs(),
        'ver_major': ver_major,
        'ver_minor': ver_minor,
        'ver_patch': ver_patch,
        'label_juju_units': config.get('label-juju-units', False),
        'alert_managers': unitdata.kv().get('alert_managers', []),
    }

    render(source=PATHS['prometheus_yml_tmpl'],
           target=PATHS['prometheus_yml'],
           context=options
           )
    validate_config()

    if config.changed('prometheus_registration_listen'):
        render(source='promreg.yaml.j2',
               target=PATHS['promreg_yml'],
               context=options,
               perms=0o440
               )
        hookenv.open_port(config['prometheus_registration_port'])
        set_state('promreg.do-restart')

    set_state('prometheus.do-reload')
    remove_state('prometheus.do-reconfig-yml')


def check_ports(new_port):
    kv = unitdata.kv()
    if kv.get('prometheus.port') != new_port:
        hookenv.open_port(new_port)
        if kv.get('prometheus.port'):  # Dont try to close non existing ports
            hookenv.close_port(kv.get('prometheus.port'))
        kv.set('prometheus.port', new_port)


@when('prometheus.do-reconfig-def')
def write_prometheus_config_def():
    hookenv.status_set('maintenance', 'Updating configuration')
    config = hookenv.config()
    port = config.get('web-listen-port', 9090)
    check_ports(port)
    kv = unitdata.kv()
    if config.get('external_url', False):
        vars = {
            'private_address': hookenv.unit_get('private-address'),
            'public_address': hookenv.unit_get('public-address'),
            'port': port,
        }
        runtime_args('--web.external-url',
                     config['external_url'].format(**vars))
    runtime_args('--web.listen-address',
                 ":{}".format(kv.get('prometheus.port')))
    args = runtime_args()
    hookenv.log('runtime_args: {}'.format(args))
    if args:
        render(source=PATHS['prometheus_def_tmpl'],
               target=PATHS['prometheus_def'],
               context={'args': args},
               )
    set_state('prometheus.do-restart')
    remove_state('prometheus.do-reconfig-def')


@when_not('prometheus.started')
def setup_prometheus():
    if is_state('prometheus.do-install'):
        hookenv.status_set('waiting',
                           'Waiting for installation to complete')
        return
    hookenv.status_set('maintenance', 'Configuring software')
    set_datadir_perms()
    install_packages()
    set_state('prometheus.do-check-reconfig')


@when('prometheus.started')
def check_config():
    set_state('prometheus.do-check-reconfig')


@when('prometheus.do-check-reconfig')
@when_not('prometheus.do-install')
def check_reconfig_prometheus():
    config = hookenv.config()
    kv = unitdata.kv()
    target_jobs = kv.get('target_jobs', [])
    scrape_jobs = kv.get('scrape_jobs', [])
    manual_jobs = kv.get('manual_jobs', [])
    alert_managers = kv.get('alert_managers', [])
    args = runtime_args()
    install_opts = ('snap_channel', 'prometheus_registration_listen')
    def_change_opts = ('daemon-args', 'external_url', 'web-listen-port')
    rules_change_opts = ('custom-rules', 'rules-repo', 'rules-update-schedule',
                         'rules-update-ssh-key', 'rules-update-notify-email')
    yml_change_opts = set(config.keys()) - set(install_opts + def_change_opts + rules_change_opts)
    yml_change_opts = yml_change_opts.union(['external_url', 'web-listen-port'])
    if any(config.changed(opt) for opt in install_opts):
        set_state('prometheus.do-install')
    if any(config.changed(opt) for opt in def_change_opts):
        hookenv.log('One of options affecting daemon args has changed, scheduling reconfiguration')
        set_state('prometheus.do-reconfig-def')
    if any(config.changed(opt) for opt in rules_change_opts) \
       or templates_changed([PATHS['refresh_rules_script_tmpl']]):
        configure_rules()
    if any(config.changed(opt) for opt in yml_change_opts):
        set_state('prometheus.do-reconfig-yml')
    if any((
        data_changed('prometheus.target_jobs', target_jobs),
        data_changed('prometheus.scrape_jobs', scrape_jobs),
        data_changed('prometheus.manual_jobs', manual_jobs),
        data_changed('prometheus.alert_managers', alert_managers),
        templates_changed([PATHS['prometheus_yml_tmpl']]),
    )):
        set_state('prometheus.do-reconfig-yml')
    if data_changed('prometheus.args', args) \
       or templates_changed([PATHS['prometheus_def_tmpl']]):
        hookenv.log('Possible CLI format changes, scheduling reconfiguration')
        set_state('prometheus.do-reconfig-def')
    # Cron can update rules on disk, update-status hook can trigger config update
    if kv.get('prometheus.rule_files') != get_rule_files():
        kv.set('prometheus.rule_files', get_rule_files())
        set_state('prometheus.do-reconfig-yml')

    remove_state('prometheus.do-check-reconfig')


@when('prometheus.do-restart')
@when_not('prometheus.do-install')
def restart_prometheus():
    if not host.service_running(PATHS['service_name']):
        hookenv.log('Starting {}...'.format(PATHS['service_name']))
        host.service_start(PATHS['service_name'])
    else:
        hookenv.log('Restarting {}, config file changed...'.format(PATHS['service_name']))
        host.service_restart(PATHS['service_name'])
        hookenv.log('Sleeping for 30s to make sure prometheus has time to start. '
                    'Needed in case of multiple configuration changes made at short intervals')
        time.sleep(30)

    hookenv.status_set('active', 'Ready')
    set_state('prometheus.started')
    remove_state('prometheus.do-restart')
    remove_state('prometheus.do-reload')  # We've just restarted, no need to reload


@when('prometheus.started')
@when('prometheus.do-reload')
@when_not('prometheus.do-install')
@when_not('prometheus.do-restart')
def reload_prometheus():
    if host.service_running(PATHS['service_name']):
        # Use SIGHUP to reload config gracefully
        subprocess.check_call(['killall', '-HUP', 'prometheus'])
        remove_state('prometheus.do-reload')
        hookenv.status_set('active', 'Ready')
    else:
        # Let's make sure prometheus is up
        set_state('prometheus.do-restart')


@when('promreg.do-restart')
def restart_promreg():
    hookenv.log('Restarting promreg')
    if not host.service_running(PATHS['promreg_service_name']):
        host.service_start(PATHS['promreg_service_name'])
    else:
        hookenv.log('Restarting {}, config file changed...'.format(PATHS['promreg_service_name']))
        host.service_restart(PATHS['promreg_service_name'])
    remove_state('promreg.do-restart')


# Relations
@when('prometheus.started')
@when_not('target.available')
def update_prometheus_no_targets():
    unitdata.kv().set('target_jobs', [])
    data_changed('target.related_services', [])
    set_state('prometheus.do-check-reconfig')


@when('prometheus.started')
@when_not('scrape.available')
def update_prometheus_no_scrapes():
    unitdata.kv().set('scrape_jobs', [])
    set_state('prometheus.do-check-reconfig')


@when('prometheus.started')
@when_not('endpoint.manual-jobs.has_jobs')
def update_prometheus_no_manual_jobs():
    unitdata.kv().set('manual_jobs', [])
    set_state('prometheus.do-check-reconfig')


@when('prometheus.started')
@when('target.available')
def update_prometheus_targets(target):
    unit_names = get_related_unit_names()
    services = target.services()
    related_targets = []
    for service in sorted(services, key=lambda k: k['service_name']):
        targets = []
        for unit in service['hosts']:
            hookenv.log('{} has a unit {}:{}'.format(
                service['service_name'],
                unit['hostname'],
                unit['port']), level='debug')
            hostname = '{hostname}:{port}'.format(**unit)
            targets.append((hostname, unit_names.get(hostname)))
        related_targets.append({'job_name': service['service_name'],
                                'targets': sorted(targets)})

    unitdata.kv().set('target_jobs', related_targets)
    set_state('prometheus.do-check-reconfig')


@when('prometheus.started')
@when('prometheus-rules.available')
def update_prometheus_rules(prometheus_rules):
    rules = prometheus_rules.prometheus_rules()
    if not data_changed('prometheus.rules', rules):  # Assumes rules is stable!
        return

    # insert into a template
    template = """
groups:
  - name: JujuRelation
    rules:
{{ rules }}
    """
    with open(PATHS['juju_rules_path'], 'w') as fh:
        fh.write(render_template(template, {'rules': "\n".join(rules)}))
    set_state('prometheus.do-check-reconfig')
    set_state('prometheus.do-restart')


def get_related_unit_names():
    '''Return a dict mapping unit hostnames to names.'''
    app_units = {}
    for relation_id in hookenv.relation_ids('target'):
        for relation_data in hookenv.relations_for_id(relation_id):
            hostname = relation_data.get('hostname')
            port = relation_data.get('port')
            if not (hostname and port):
                continue
            hostname = '{}:{}'.format(hostname, port)
            unit = relation_data['__unit__']
            # The name of the principal unit in case the remote end is a
            # subordinate charm.
            principal_unit = hookenv.relation_get(
                'principal-unit', unit, relation_id)
            app_units[hostname] = principal_unit or unit
    return app_units


@when('prometheus.started')
@when('scrape.available')
def update_prometheus_scrape_targets(target):
    targets = target.targets()
    unitdata.kv().set('scrape_jobs', targets)
    set_state('prometheus.do-check-reconfig')


@when('endpoint.manual-jobs.has_jobs')
def update_prometheus_manual_jobs(manual_jobs):
    job_jsons = []
    certs_dir = Path(PATHS['certs_dir'])
    certs_dir.mkdir(parents=True, exist_ok=True)
    for job in manual_jobs.jobs:
        if job.ca_cert:
            # escape path special chars to be safe when saving
            job_name = job.job_name.replace('/', '_')
            request_id = job.request_id.replace('/', '_')
            # escape glob special chars to be safe when cleaning up
            g_job_name = glob.escape(job_name)
            g_request_id = glob.escape(request_id)
            # include a hash of the cert data rather than just the request ID
            # so that check_reconfig_prometheus can detect changes to the data
            # via the filename
            cert_hash = sha1(job.ca_cert.encode('utf8')).hexdigest()
            cert_path = certs_dir / '{}-{}-{}.crt'.format(job_name,
                                                          request_id,
                                                          cert_hash)
            # avoid re-writing the file if contents haven't changed
            if not cert_path.exists():
                # clean up previous versions of the cert
                for old_cert in certs_dir.glob('{}-{}-*.crt'
                                               ''.format(g_job_name,
                                                         g_request_id)):
                    old_cert.unlink()
                # NB: ensure trailing newline because some programs fail
                # without it (haven't tested prometheus specifically)
                cert_path.write_text(job.ca_cert.rstrip() + '\n')
            job_jsons.append(job.to_json(cert_path))
        else:
            job_jsons.append(job.to_json())
    unitdata.kv().set('manual_jobs', job_jsons)
    set_state('prometheus.do-check-reconfig')


@when('prometheus.started')
@when_not('alertmanager-service.available')
def update_prometheus_no_alertmanager():
    # XXX switch to config file
    unitdata.kv().unset('alert_managers')
    set_state('prometheus.do-check-reconfig')


@when('prometheus.started')
@when('alertmanager-service.available')
def update_prometheus_alertmanager(alertmanager):
    services = alertmanager.services()
    if not (data_changed('alertmanager-service.related_services', services)):
        return
    alert_managers = []
    for service in services:
        for unit in service['hosts']:
            hookenv.log('{} has a unit {}:{}'.format(service['service_name'], unit['hostname'], unit['port']))
            alert_managers.append('{hostname}:{port}'.format(**unit))
    unitdata.kv().set('alert_managers', sorted(alert_managers))
    set_state('prometheus.do-check-reconfig')


@when('prometheus.started')
@when_not('snmp-exporter.available')
def update_prometheus_no_snmp_exporter():
    unitdata.kv().unset('snmp_exporter_address')
    set_state('prometheus.do-check-reconfig')


@when('prometheus.started')
@when_not('blackbox-exporter.available')
def update_prometheus_no_blackbox_exporter():
    unitdata.kv().unset('blackbox_exporter_address')
    set_state('prometheus.do-check-reconfig')


@when('prometheus.started')
@when('snmp-exporter.available')
def update_prometheus_snmp_exporter(snmp_exporter):
    services = snmp_exporter.services()
    if not (data_changed('snmp-exporter.related_services', services)):
        return
    # At this point we only support single snmp exporter target
    for service in services:
        for unit in service['hosts']:
            hookenv.log('{} has a unit {}:{}'.format(
                        service['service_name'], unit['hostname'], unit['port']))
            unitdata.kv().set('snmp_exporter_address', '{hostname}:{port}'.format(**unit))
    set_state('prometheus.do-check-reconfig')


@when('prometheus.started')
@when('blackbox-exporter.available')
def update_prometheus_blackbox_exporter(blackbox_exporter):
    services = blackbox_exporter.services()
    if not (data_changed('blackbox-exporter.related_services', services)):
        return
    # At this point we only support single blackbox exporter target
    for service in services:
        for unit in service['hosts']:
            hookenv.log('{} has a unit {}:{}'.format(
                        service['service_name'], unit['hostname'], unit['port']))
            unitdata.kv().set('blackbox_exporter_address', '{hostname}:{port}'.format(**unit))
    set_state('prometheus.do-check-reconfig')


@when('prometheus.started')
@when('nrpe-external-master.available')
def update_nrpe_config(svc):
    hostname = nrpe.get_nagios_hostname()
    nrpe_setup = nrpe.NRPE(hostname=hostname)
    port = unitdata.kv().get('prometheus.port', 9090)
    nrpe_setup.add_check('prometheus_http',
                         'Prometheus HTTP check',
                         'check_http -I 127.0.0.1 -p {} -u /graph'.format(port)
                         )
    nrpe_setup.write()


@when('prometheus.started')
@when('grafana-source.available')
def provide_grafana_source(grafana):
    kv = unitdata.kv()
    port = kv.get('prometheus.port')
    grafana.provide('prometheus', port, 'Juju generated source')


@when('prometheus.started')
@when('website.available')
def configure_website(website):
    port = unitdata.kv().get('prometheus.port', 9090)
    website.configure(port=port)
